/*
 *	Copyright (C) 2007, 2008, 2009, 2010. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file
@ingroup OpenCVExamplePrograms
@brief Obtains the Delaunay triangulation for a set of points.

This application detects a set of feature points on each frame of the input video sequence, and shows the corresponding <a href="http://en.wikipedia.org/wiki/Delaunay_triangulation">Delaunay triangulation</a> on an image canvas.

An example plot of that triangulation can be seen in the following image:

@image html opencv-delaunay.png

This application uses OpenCV functionality to obtain the triangulation.

@section Usage
Once compiled, the program should be used with this command line:
\code
	./delaunay --URL=<path or url to video>
\endcode

@param URL Video source identificator. See section @ref QVMPlayerURLFormat for further info about its format.

@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>

#include <QDebug>
#include <QThread>

#include <QVApplication>
#include <QVMPlayerReaderBlock>
#include <QVDefaultGUI>
#include <QVImageCanvas>
#include <QVPolylineF>

#include <QVHarrisPointDetector>
#include <qvip.h>
#include <qvimageio.h>

#ifndef DOXYGEN_IGNORE_THIS

#ifdef OPENCV
#include <QVDelaunayTriangulation>

QHash<QPointF, QPointF> openCVDelaunayTriangulation(const QList<QPointF> &selectedPoints)
	{
	// STORAGE AND STRUCTURE FOR DELAUNAY SUBDIVISION
	/*CvMemStorage*	storage = cvCreateMemStorage(0);		//Storage for the Delaunay subdivsion

	// The subdivision itself
	CvSubdiv2D* subdiv = cvCreateSubdiv2D(	CV_SEQ_KIND_SUBDIV2D,
						sizeof(*subdiv),
						sizeof(CvSubdiv2DPoint),
						sizeof(CvQuadEdge2D),
						storage);*/

	// Set boundaries for the Delaunay triangulation algorithm.
	double	minX = std::numeric_limits<double>::max(),
		minY = std::numeric_limits<double>::max(),
		maxX = std::numeric_limits<double>::min(),
		maxY = std::numeric_limits<double>::min();

	foreach(QPointF point, selectedPoints)
		{
		minX = MIN(minX, point.x());
		minY = MIN(minY, point.y());
		maxX = MAX(maxX, point.x());
		maxY = MAX(maxY, point.y());
		}

	QVDelaunayTriangulation delaunayTriangulation(QRect(minX, minY, maxX - minX, maxY - minY));

	//cvInitSubdivDelaunay2D(subdiv, (CvRect){ minX, minY, maxX, maxY });

	// Add points to the Delaunay subdivision structure.
	foreach(QPointF point, selectedPoints)
		//{
		//cvSubdivDelaunay2DInsert(subdiv, (CvPoint2D32f){ point.x(), point.y() });
		delaunayTriangulation.addPoint(point);
		//}
	return delaunayTriangulation.getLinks();

	// Transverse Delaunay partition, adding Delaunay triangulation edges.
	/*CvSeqReader reader;
	QHash <QPointF, QPointF> correspondences;
	cvStartReadSeq((CvSeq *)subdiv->edges, &reader);

	for(int i=0; i<subdiv->edges->total; i++)
		{
		CvQuadEdge2D * edge = (CvQuadEdge2D *) reader.ptr;
		if(CV_IS_SET_ELEM(edge))
			{
			CvSubdiv2DPoint * org = cvSubdiv2DEdgeOrg( (CvSubdiv2DEdge) edge );
			CvSubdiv2DPoint * dst = cvSubdiv2DEdgeDst( (CvSubdiv2DEdge) edge );
			if(org && dst && (CV_IS_SET_ELEM(org) && ((org)->flags & CV_SUBDIV2D_VIRTUAL_POINT_FLAG)==0))
				if ( (org->pt.x > minX) && (org->pt.x < maxX) && (dst->pt.y > minY) && (dst->pt.y < maxY) )
					correspondences.insertMulti(QPointF(org->pt.x, org->pt.y), QPointF(dst->pt.x, dst->pt.y));
			}
		CV_NEXT_SEQ_ELEM(subdiv->edges->elem_size, reader);
		}*/

	//const QHash <QPointF, QPointF> correspondences2 = delaunayTriangulation.getLinks();

	/*foreach(QPointF p, correspondences.keys())
		if (correspondences.values(p).toSet() != correspondences2.values(p).toSet())
			std::cout << "ERROR" << std::endl;			*/

	//return correspondences2;
	}

class DelaunayBlock: public QVProcessingBlock
	{
	private:
		QList<QVPolylineF> map2polyline(const QHash<QPointF, QPointF> &map)
			{
			QList<QVPolylineF> polylines;
		
			QHashIterator<QPointF, QPointF> iterator(map);
		
			while (iterator.hasNext())
				{
				iterator.next();
				QVPolylineF polyline;
				polyline.append(iterator.key());
				polyline.append(iterator.value());
				polylines.append(polyline);
				}
		
			return polylines;
			}

	public:
		DelaunayBlock(const QString name): QVProcessingBlock(name)
			{
			addProperty< QList < QPointF > >("Feature locations", inputFlag|outputFlag);
			addProperty< QList<QVPolylineF> >("Delaunay edges", outputFlag);
			}

		void iterate()
			{
			// 0. Read input property values.
			const QList<QPointF> selectedPoints = getPropertyValue< QList< QPointF > >("Feature locations");
			timeFlag("Read data");

			// 1. Get Delaunay triangulation.
			const QHash<QPointF, QPointF> correspondences = openCVDelaunayTriangulation(selectedPoints);
			timeFlag("Generated edges");

			// 2. Generate the graph as a list of polylines for visualization.
			const QList<QVPolylineF> delaunayEdges = map2polyline(correspondences);
			timeFlag("From correspondences to polylines");

			setPropertyValue< QList<QVPolylineF> >("Delaunay edges", delaunayEdges);
			}
	};

#include <QVYUV4MPEG2WriterBlock>
#include <QVNumericPlot>
int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		"Example program for QVision library. Obtains intrinsic and extrinsic camera parameters."
		);

	QVMPlayerReaderBlock camera("Video");

	QVHarrisPointDetector pointDetector("Harris corners detector");
	camera.linkProperty(&pointDetector,"Input image");

	// Camera calibrator.
	DelaunayBlock delaunayBlock("Delaunay");
	pointDetector.linkProperty("Feature locations", delaunayBlock, QVProcessingBlock::SynchronousLink);

	// Image outputs
	QVImageCanvas imageCanvas("Delaunay regions");
	pointDetector.linkProperty("Input image", imageCanvas);
	delaunayBlock.linkProperty("Feature locations", imageCanvas);
	delaunayBlock.linkProperty("Delaunay edges", imageCanvas);
	imageCanvas.setColor("Delaunay edges", Qt::red);

	QVDefaultGUI interface;

	return app.exec();
	}
#else
int main(int argc, char *argv[])
	{
	std::cout << "ERROR: OpenCV compatibility was not activated in QVision compilation." << std::endl;
	return -1;
	}
#endif	// OPENCV
#endif	// DOXIGEN_IGNORE_THIS

